@@ -117,7 +117,7 @@
 /*---------------------------------------------------------------------*/
 /* VAL_STACK_SIZE: The size of the operand stack for expr. parsing     */
 /*---------------------------------------------------------------------*/
-#define VAL_STACK_SIZE 1000
+#define VAL_STACK_SIZE 100
 
 /*---------------------------------------------------------------------*/
 /* INCLUDE_NEST: How many nested INCLUDES do we handle?                */
@@ -166,6 +166,10 @@
 
 #define PASSTHRU_LEN 32
 
+#define MAX_RECURSION_LEVEL 1000
+
+#define MAX_FUNC_ARGS 64
+
 #define PSBEGIN "# rem2ps begin"
 #define PSEND   "# rem2ps end"
 
