@@ -23,9 +23,8 @@
 #include "globals.h"
 #include "err.h"
 #include "protos.h"
-#include "expr.h"
 
-static int ParseTimeTrig (ParsePtr s, TimeTrig *tim, int save_in_globals);
+static int ParseTimeTrig (ParsePtr s, TimeTrig *tim);
 static int ParseLocalOmit (ParsePtr s, Trigger *t);
 static int ParseScanFrom (ParsePtr s, Trigger *t, int type);
 static int ParsePriority (ParsePtr s, Trigger *t);
@@ -63,7 +62,7 @@ int DoRem(ParsePtr p)
     DBufInit(&buf);
 
     /* Parse the trigger date and time */
-    if ( (r=ParseRem(p, &trig, &tim, 1)) ) {
+    if ( (r=ParseRem(p, &trig, &tim)) ) {
 	FreeTrig(&trig);
 	return r;
     }
@@ -163,6 +162,7 @@ int DoRem(ParsePtr p)
 	    if (p->expr_happened) {
 		if (p->nonconst_expr) {
 		    PurgeEchoLine("%s\n", "#!P: Next line may have expired, but contains non-constant expression");
+		    PurgeEchoLine("%s\n", "#!P: or a relative SCANFROM clause");
 		    PurgeEchoLine("%s\n", CurLine);
 		} else {
 		    PurgeEchoLine("%s\n", "#!P: Next line has expired, but contains expression...  please verify");
@@ -220,7 +220,7 @@ int DoRem(ParsePtr p)
 /*  trigger structure.                                         */
 /*                                                             */
 /***************************************************************/
-int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
+int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim)
 {
     register int r;
     DynamicBuffer buf;
@@ -262,10 +262,6 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
     trig->need_wkday = 0;
     trig->adj_for_last = 0;
 
-    if (save_in_globals) {
-	LastTriggerTime = NO_TIME;
-    }
-
     int parsing = 1;
     while(parsing) {
 	/* Read space-delimited string */
@@ -315,10 +311,6 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
 	    trig->m = m;
 	    trig->d = d;
 	    tim->ttime = (tok.val % MINUTES_PER_DAY);
-	    if (save_in_globals) {
-		LastTriggerTime = tim->ttime;
-		SaveLastTimeTrig(tim);
-	    }
 	    break;
 
 	case T_WkDay:
@@ -355,14 +347,14 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
             DBufFree(&buf);
             if (tim->ttime != NO_TIME) return E_TIME_TWICE;
             tim->ttime = tok.val;
-            r = ParseTimeTrig(s, tim, save_in_globals);
+            r = ParseTimeTrig(s, tim);
             if (r) return r;
 	    trig->duration_days = ComputeTrigDuration(tim);
 	    break;
 
 	case T_At:
 	    DBufFree(&buf);
-	    r=ParseTimeTrig(s, tim, save_in_globals);
+	    r=ParseTimeTrig(s, tim);
 	    if (r) return r;
 	    trig->duration_days = ComputeTrigDuration(tim);
 	    break;
@@ -481,10 +473,10 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
 	    r=ParseToken(s, &buf);
 	    if (r) return r;
 	    StrnCpy(trig->omitfunc, DBufValue(&buf), VAR_NAME_LEN);
-
+            strtolower(trig->omitfunc);
 	    /* An OMITFUNC counts as a nonconst_expr! */
-	    s->expr_happened = 1;
-	    s->nonconst_expr = 1;
+            s->expr_happened = 1;
+            s->nonconst_expr = 1;
 	    DBufFree(&buf);
 	    break;
 
@@ -492,6 +484,7 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
 	    r=ParseToken(s, &buf);
 	    if(r) return r;
 	    StrnCpy(trig->warn, DBufValue(&buf), VAR_NAME_LEN);
+            strtolower(trig->warn);
 	    DBufFree(&buf);
 	    break;
 
@@ -522,9 +515,6 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
 		} else {
 		    tim->duration = NO_TIME;
 		}
-		if (save_in_globals) {
-		    SaveLastTimeTrig(tim);
-		}
 		trig->duration_days = ComputeTrigDuration(tim);
 		break;
 	    default:
@@ -536,6 +526,7 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
 	    r=ParseToken(s, &buf);
 	    if(r) return r;
 	    StrnCpy(trig->sched, DBufValue(&buf), VAR_NAME_LEN);
+            strtolower(trig->sched);
 	    DBufFree(&buf);
 	    break;
 
@@ -608,7 +599,7 @@ int ParseRem(ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals)
 /*  ParseTimeTrig - parse the AT part of a timed reminder      */
 /*                                                             */
 /***************************************************************/
-static int ParseTimeTrig(ParsePtr s, TimeTrig *tim, int save_in_globals)
+static int ParseTimeTrig(ParsePtr s, TimeTrig *tim)
 {
     Token tok;
     int r;
@@ -643,11 +634,6 @@ static int ParseTimeTrig(ParsePtr s, TimeTrig *tim, int save_in_globals)
 	default:
 	    if (tim->ttime == NO_TIME) return E_EXPECT_TIME;
 
-	    /* Save trigger time in global variable */
-	    if (save_in_globals) {
-		LastTriggerTime = tim->ttime;
-		SaveLastTimeTrig(tim);
-	    }
 	    PushToken(DBufValue(&buf), s);
 	    DBufFree(&buf);
 	    return OK;
@@ -679,6 +665,9 @@ static int ParseLocalOmit(ParsePtr s, Trigger *t)
 	    break;
 
 	default:
+            if (t->localomit == NO_WD) {
+                return E_EXPECTING_WEEKDAY;
+            }
 	    PushToken(DBufValue(&buf), s);
 	    DBufFree(&buf);
 	    return OK;
@@ -873,6 +862,9 @@ static int ParseScanFrom(ParsePtr s, Trigger *t, int type)
 		tok.val = -tok.val;
 	    }
 	    FromDSE(DSEToday - tok.val, &y, &m, &d);
+            /* Don't purge reminders with a relative scanfrom */
+            s->expr_happened = 1;
+            s->nonconst_expr = 1;
 	    break;
 
 	default:
@@ -1281,20 +1273,28 @@ int DoSatRemind(Trigger *trig, TimeTrig *tt, ParsePtr p)
 {
     int iter, dse, r, start;
     Value v;
-    char const *s;
-    char const *t;
+    expr_node *sat_node;
+    int nonconst = 0;
 
-    t = p->pos;
+    sat_node = ParseExpr(p, &r);
+    if (r != OK) {
+        return r;
+    }
+    if (!sat_node) {
+        return E_SWERR;
+    }
     iter = 0;
     start = trig->scanfrom;
     while (iter++ < MaxSatIter) {
 	dse = ComputeTriggerNoAdjustDuration(start, trig, tt, &r, 1, 0);
 	if (r) {
+            free_expr_tree(sat_node);
 	    if (r == E_CANT_TRIG) return OK; else return r;
 	}
 	if (dse != start && trig->duration_days) {
 	    dse = ComputeTriggerNoAdjustDuration(start, trig, tt, &r, 1, trig->duration_days);
 	    if (r) {
+                free_expr_tree(sat_node);
 		if (r == E_CANT_TRIG) return OK; else return r;
 	    }
 	} else if (dse == start) {
@@ -1307,13 +1307,18 @@ int DoSatRemind(Trigger *trig, TimeTrig *tt, ParsePtr p)
 	    SaveAllTriggerInfo(trig, tt, dse, tt->ttime, 1);
 	}
 	if (dse == -1) {
+            free_expr_tree(sat_node);
 	    return E_EXPIRED;
 	}
-	s = p->pos;
-	r = EvaluateExpr(p, &v);
-	t = p->pos;
-	if (r) return r;
-	if (v.type != INT_TYPE && v.type != STR_TYPE) return E_BAD_TYPE;
+        r = evaluate_expression(sat_node, NULL, &v, &nonconst);
+	if (r) {
+            free_expr_tree(sat_node);
+            return r;
+        }
+	if (v.type != INT_TYPE && v.type != STR_TYPE) {
+            free_expr_tree(sat_node);
+            return E_BAD_TYPE;
+        }
 	if ((v.type == INT_TYPE && v.v.val) ||
 	    (v.type == STR_TYPE && *v.v.str)) {
 	    AdjustTriggerForDuration(trig->scanfrom, dse, trig, tt, 1);
@@ -1338,17 +1343,17 @@ int DoSatRemind(Trigger *trig, TimeTrig *tt, ParsePtr p)
 		}
 		fprintf(ErrFp, "\n");
 	    }
+            free_expr_tree(sat_node);
 	    return OK;
 	}
-	p->pos = s;
 	if (dse+trig->duration_days < start) {
 	    start++;
 	} else {
 	    start = dse+trig->duration_days+1;
 	}
     }
-    p->pos = t;
     LastTrigValid = 0;
+    free_expr_tree(sat_node);
     return E_CANT_TRIG;
 }
 
