@@ -12,7 +12,7 @@
 /***************************************************************/
 
 #include "config.h"
-#include "expr.h"
+#include "types.h"
 #define L_IN_DOSUBST
 #include <stdio.h>
 #include <string.h>
@@ -20,7 +20,6 @@
 
 #include <stdlib.h>
 
-#include "types.h"
 #include "globals.h"
 #include "err.h"
 #include "protos.h"
@@ -50,7 +49,7 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
     int err, done;
     int c;
     int d, m, y;
-    int tim = tt->ttime;
+    int tim = NO_TIME;
     int h, min, hh, ch, cmin, chh;
     int i;
     char const *pm, *cpm;
@@ -72,6 +71,9 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
 
     FromDSE(dse, &y, &m, &d);
 
+    if (tt) {
+        tim = tt->ttime;
+    }
     if (tim == NO_TIME) tim = curtime;
     tdiff = tim - curtime;
     adiff = ABS(tdiff);
@@ -139,6 +141,7 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
             } else {
                 r = -1;
             }
+            DestroyValue(v);
         } else {
             Eprint("%s", ErrMsg[r]);
         }
@@ -163,6 +166,7 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
             } else {
                 r = -1;
             }
+            DestroyValue(v);
         } else {
             Eprint("%s", ErrMsg[r]);
         }
@@ -239,7 +243,7 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
                     break;
                 }
                 if (i < 64) {
-                    *ss++ = c;
+                    *ss++ = tolower(c);
                     *ss = 0;
                     i++;
                 }
@@ -266,10 +270,10 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
             continue;
         }
 	done = 0;
-        snprintf(uf, sizeof(uf), "subst_%c", c);
+        snprintf(uf, sizeof(uf), "subst_%c", tolower(c));
         if (UserFuncExists(uf) == 3) {
             snprintf(s, sizeof(s), "subst_%c(%d,'%04d-%02d-%02d',%02d:%02d)",
-                     c, altmode ? 1 : 0, y, m+1, d, h, min);
+                     tolower(c), altmode ? 1 : 0, y, m+1, d, h, min);
             expr = (char const *) s;
             r = EvalExpr(&expr, &v, NULL);
             if (r == OK) {
@@ -341,10 +345,10 @@ int DoSubst(ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse,
 
 
         if (!done) {
-            snprintf(uf, sizeof(uf), "subst_%cx", c);
+            snprintf(uf, sizeof(uf), "subst_%cx", tolower(c));
             if (UserFuncExists(uf) == 3) {
                 snprintf(s, sizeof(s), "subst_%cx(%d,'%04d-%02d-%02d',%02d:%02d)",
-                         c, altmode ? 1 : 0, y, m+1, d, h, min);
+                         tolower(c), altmode ? 1 : 0, y, m+1, d, h, min);
                 expr = (char const *) s;
                 r = EvalExpr(&expr, &v, NULL);
                 if (r == OK) {
