@@ -23,6 +23,7 @@
 #define INIT(var, val) var
 #endif
 
+#include <signal.h>
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -39,7 +40,7 @@ EXTERN  FILE *ErrFp;
 #define IsLeapYear(y) (((y) % 4) ? 0 : ((!((y) % 100) && ((y) % 400)) ? 0 : 1 ))
 #define DaysInMonth(m, y) ((m) != 1 ? MonthDays[m] : 28 + IsLeapYear(y))
 
-#define DestroyValue(x) (void) (((x).type == STR_TYPE && (x).v.str) ? (free((x).v.str),(x).v.str = NULL,(x).type = ERR_TYPE) : 0)
+#define DestroyValue(x) do { if ((x).type == STR_TYPE && (x).v.str) { free((x).v.str); (x).v.str = NULL; } (x).type = ERR_TYPE; } while (0)
 
 EXTERN	int	DSEToday;
 EXTERN	int	RealToday;
@@ -73,6 +74,9 @@ EXTERN  INIT(	int 	InfiniteDelta, 0);
 EXTERN  INIT(   int     DefaultTDelta, 0);
 EXTERN  INIT(   int     DeltaOverride, 0);
 EXTERN  INIT(   int     RunDisabled, 0);
+EXTERN  INIT(   int     ExpressionEvaluationDisabled, 0);
+EXTERN  INIT(   int     ExpressionEvaluationTimeLimit, 0);
+EXTERN  INIT(   volatile sig_atomic_t ExpressionTimeLimitExceeded, 0);
 EXTERN  INIT(   int     IgnoreOnce, 0);
 EXTERN  INIT(   int     SortByTime, 0);
 EXTERN  INIT(   int     SortByDate, 0);
@@ -110,11 +114,12 @@ EXTERN  INIT(   int     PurgeIncludeDepth, 0);
 EXTERN  INIT(   FILE    *PurgeFP,  NULL);
 EXTERN  INIT(   int     NumIfs,    0);
 EXTERN  INIT(   unsigned int IfFlags,   0);
+EXTERN  INIT(   int IfLinenos[IF_NEST], {0});
 EXTERN  INIT(   int     LastTrigValid, 0);
 EXTERN  Trigger  LastTrigger;
 EXTERN  TimeTrig LastTimeTrig;
 EXTERN  INIT(   int     LastTriggerDate, 0);
-EXTERN  INIT(   int     LastTriggerTime, 0);
+EXTERN  INIT(   int     LastTriggerTime, NO_TIME);
 EXTERN  INIT(   int     ShouldCache, 0);
 EXTERN  char const   *CurLine;
 EXTERN  INIT(   int     NumTriggered, 0);
@@ -159,6 +164,9 @@ EXTERN DynamicBuffer Banner;
 EXTERN DynamicBuffer LineBuffer;
 EXTERN DynamicBuffer ExprBuf;
 
+/* User-func recursion level */
+EXTERN  INIT(   unsigned int FuncRecursionLevel, 0);
+
 extern int NumFullOmits, NumPartialOmits;
 
 /* List of months */
