@@ -19,7 +19,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
-
+#include <time.h>
 #include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -38,7 +38,6 @@
 #include "types.h"
 #include "globals.h"
 #include "protos.h"
-#include "expr.h"
 #include "err.h"
 
 static int should_guess_terminal_background = 1;
@@ -78,6 +77,7 @@ static void ProcessLongOption(char const *arg);
  *             t = Display trigger dates
  *             v = Dump variables at end
  *             l = Display entire line in error messages
+ *             s = Display expression-parsing stack usage before exit
  *  -e       = Send messages normally sent to stderr to stdout instead
  *  -z[n]    = Daemon mode waking up every n (def 1) minutes.
  *  -bn      = Time format for cal (0, 1, or 2)
@@ -604,6 +604,7 @@ void InitRemind(int argc, char const *argv[])
 	    case 'D':
 		while (*arg) {
 		    switch(*arg++) {
+                    case 's': case 'S': DebugFlag |= DB_PARSE_EXPR;  break;
 		    case 'e': case 'E': DebugFlag |= DB_ECHO_LINE;   break;
 		    case 'x': case 'X': DebugFlag |= DB_PRTEXPR;     break;
 		    case 't': case 'T': DebugFlag |= DB_PRTTRIG;     break;
@@ -804,6 +805,7 @@ void Usage(void)
     fprintf(ErrFp, " -m     Start calendar with Monday rather than Sunday\n");
     fprintf(ErrFp, " -y     Synthesize tags for tagless reminders\n");
     fprintf(ErrFp, " -j[n]  Run in 'purge' mode.  [n = INCLUDE depth]\n");
+    fprintf(ErrFp, "\nRemind home page: %s\n", PACKAGE_URL);
     exit(EXIT_FAILURE);
 }
 #endif /* L_USAGE_OVERRIDE */
@@ -923,7 +925,12 @@ static void InitializeVar(char const *str)
     r = 0;
     while (*str && *str != '=') {
 	if (r < VAR_NAME_LEN) {
-	    varname[r++] = *str;
+            if (isalpha(*str) || *str == '_' || (r > 0 && *str == '(') || (r == 0 && *str == '$') || (r > 0 && isdigit(*str))) {
+                varname[r++] = *str;
+            } else {
+                fprintf(ErrFp, ErrMsg[M_I_OPTION], ErrMsg[E_ILLEGAL_CHAR]);
+                return;
+            }
 	}
 	if (*str == '(') {
 	    /* Do a function definition if we see a paren */
@@ -933,10 +940,28 @@ static void InitializeVar(char const *str)
 	str++;
     }
     varname[r] = 0;
-    if (!*str) {
-	fprintf(ErrFp, ErrMsg[M_I_OPTION], ErrMsg[E_MISS_EQ]);
+    if (!*varname) {
+	fprintf(ErrFp, ErrMsg[M_I_OPTION], ErrMsg[E_MISS_VAR]);
 	return;
     }
+    if (!*str) {
+        /* Setting a system var does require =expr on the commandline */
+        if (*varname == '$') {
+            fprintf(ErrFp, ErrMsg[M_I_OPTION], ErrMsg[E_MISS_EQ]);
+            return;
+        }
+        val.type = INT_TYPE;
+        val.v.val = 0;
+        r = SetVar(varname, &val);
+        if (!r) {
+            r = PreserveVar(varname);
+        }
+        if (r) {
+            fprintf(ErrFp, ErrMsg[M_I_OPTION], ErrMsg[r]);
+        }
+        return;
+    }
+
     if (!*varname) {
 	fprintf(ErrFp, ErrMsg[M_I_OPTION], ErrMsg[E_MISS_VAR]);
 	return;
@@ -989,13 +1014,65 @@ AddTrustedUser(char const *username)
     NumTrustedUsers++;
 }
 
+static pid_t LimiterPid = (pid_t) -1;
+
+void unlimit_execution_time(void)
+{
+    if (LimiterPid != (pid_t) -1) {
+        kill(LimiterPid, SIGTERM);
+        LimiterPid = (pid_t) -1;
+    }
+}
+
+static void limit_execution_time(int t)
+{
+    pid_t parent = getpid();
+
+    pid_t pid = fork();
+    if (pid < 0) {
+        perror("fork");
+        exit(1);
+    }
+
+    if (pid > 0) {
+        LimiterPid = pid;
+        /* In the parent */
+        return;
+    }
+
+    /* In the child */
+    time_t start = time(NULL);
+    while(1) {
+        sleep(1);
+        if (kill(parent, 0) < 0) {
+            /* Parent has probably exited */
+            exit(0);
+        }
+        if (time(NULL) - start > t) {
+            kill(parent, SIGXCPU);
+            exit(0);
+        }
+    }
+}
+
 static void
 ProcessLongOption(char const *arg)
 {
+    int t;
     if (!strcmp(arg, "version")) {
         printf("%s\n", VERSION);
         exit(EXIT_SUCCESS);
     }
+    if (sscanf(arg, "max-execution-time=%d", &t) == 1) {
+        if (t < 0) {
+            fprintf(ErrFp, "%s: --max-execution-time must be non-negative\n", ArgV[0]);
+            return;
+        }
+        if (t > 0) {
+            limit_execution_time(t);
+        }
+        return;
+    }
     fprintf(ErrFp, "%s: Unknown long option --%s\n", ArgV[0], arg);
 }
 
@@ -1004,7 +1081,7 @@ guess_terminal_background(int *r, int *g, int *b)
 {
     int ttyfd;
     struct pollfd p;
-    int rr, gg, bb;
+    unsigned int rr, gg, bb;
     char buf[128];
     int n;
 
@@ -1070,9 +1147,9 @@ guess_terminal_background(int *r, int *g, int *b)
         /* Couldn't scan color codes */
         return;
     }
-    *r = (rr >> 8) & 255;
-    *g = (gg >> 8) & 255;
-    *b = (bb >> 8) & 255;
+    *r = (int) ((rr >> 8) & 255);
+    *g = (int) ((gg >> 8) & 255);
+    *b = (int) ((bb >> 8) & 255);
 }
 
 static struct termios orig_termios;
