@@ -22,6 +22,11 @@
 #include <stdio.h>
 #include <signal.h>
 #include <string.h>
+
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+
 #include <stdarg.h>
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -41,7 +46,6 @@
 
 #include "types.h"
 #include "protos.h"
-#include "expr.h"
 #include "globals.h"
 #include "err.h"
 
@@ -50,6 +54,38 @@ static void DoReminders(void);
 /* Macro for simplifying common block so as not to litter code */
 #define OUTPUT(c) do { if (output) { DBufPutc(output, c); } else { putchar(c); } } while(0)
 
+void
+exitfunc(void)
+{
+    /* Kill any execution-time-limiter process */
+    unlimit_execution_time();
+
+    if (DebugFlag & DB_PARSE_EXPR) {
+        fflush(stdout);
+        fflush(stderr);
+        UnsetAllUserFuncs();
+        print_expr_nodes_stats();
+    }
+}
+
+static void sigalrm(int sig)
+{
+    UNUSED(sig);
+    if (ExpressionEvaluationTimeLimit) {
+        ExpressionTimeLimitExceeded = 1;
+    }
+}
+
+static void sigxcpu(int sig)
+{
+
+    UNUSED(sig);
+    int r = write(STDERR_FILENO, "\n\nmax-execution-time exceeded.\n\n", 32);
+
+    /* Pretend to use r to avoid compiler warning */
+    _exit(1 + (r-r));
+}
+
 /***************************************************************/
 /***************************************************************/
 /**                                                           **/
@@ -61,6 +97,8 @@ int main(int argc, char *argv[])
 {
     int pid;
 
+    struct sigaction act;
+
 #ifdef HAVE_SETLOCALE
     setlocale(LC_ALL, "");
 #endif
@@ -73,9 +111,30 @@ int main(int argc, char *argv[])
     ArgV = (char const **) argv;
 
     InitRemind(argc, (char const **) argv);
+
+    act.sa_handler = sigalrm;
+    sigemptyset(&act.sa_mask);
+    act.sa_flags = SA_RESTART;
+    if (sigaction(SIGALRM, &act, NULL) < 0) {
+        fprintf(stderr, "%s: sigaction() failed: %s\n",
+                argv[0], strerror(errno));
+        exit(1);
+    }
+
+    act.sa_handler = sigxcpu;
+    act.sa_flags = SA_RESTART;
+    sigemptyset(&act.sa_mask);
+    if (sigaction(SIGXCPU, &act, NULL) < 0) {
+        fprintf(stderr, "%s: sigaction() failed: %s\n",
+                argv[0], strerror(errno));
+        exit(1);
+    }
+
     DBufInit(&(LastTrigger.tags));
     ClearLastTriggers();
 
+    atexit(exitfunc);
+
     if (DoCalendar || (DoSimpleCalendar && (!NextMode || PsCal))) {
 	ProduceCalendar();
 	return 0;
@@ -99,7 +158,7 @@ int main(int argc, char *argv[])
 	}
 
 	if (!Hush) {
-	    if (DestroyOmitContexts())
+	    if (DestroyOmitContexts(1))
 		Eprint("%s", ErrMsg[E_PUSH_NOPOP]);
 	    if (!Daemon && !NextMode && !NumTriggered && !NumQueued) {
 		printf("%s\n", ErrMsg[E_NOREMINDERS]);
@@ -152,7 +211,7 @@ void
 PerIterationInit(void)
 {
     ClearGlobalOmits();
-    DestroyOmitContexts();
+    DestroyOmitContexts(1);
     DestroyVars(0);
     DefaultColorR = -1;
     DefaultColorG = -1;
@@ -279,9 +338,11 @@ static void DoReminders(void)
 	    case T_Pop:     r=PopOmitContext(&p);     break;
 	    case T_Preserve: r=DoPreserve(&p);  break;
 	    case T_Push:    r=PushOmitContext(&p);    break;
+            case T_Expr: r = DoExpr(&p); break;
 	    case T_RemType: if (tok.val == RUN_TYPE) {
 		    r=DoRun(&p);
 		} else {
+		    DestroyParser(&p);
 		    CreateParser(CurLine, &p);
 		    r=DoRem(&p);
 		    purge_handled = 1;
@@ -290,10 +351,8 @@ static void DoReminders(void)
 
 
 	    /* If we don't recognize the command, do a REM by default */
-	    /* Note:  Since the parser hasn't been used yet, we don't */
-	    /* need to destroy it here. */
 
-	    default: CreateParser(CurLine, &p); purge_handled = 1; r=DoRem(&p); break;
+	    default: DestroyParser(&p); CreateParser(CurLine, &p); purge_handled = 1; r=DoRem(&p); break;
 
 	    }
 	    if (r && (!Hush || r != E_RUN_DISABLED)) {
@@ -451,6 +510,16 @@ int ParseChar(ParsePtr p, int *err, int peek)
 		return *(p->pos++);
 	    }
 	}
+
+        /* Convert [[ to just a literal [ */
+        if (*p->pos == BEG_OF_EXPR && *(p->pos+1) == BEG_OF_EXPR) {
+	    if (peek) {
+		return *(p->pos+1);
+	    } else {
+                p->pos++;
+		return *(p->pos++);
+	    }
+        }
 	p->expr_happened = 1;
 	p->pos++;
 	r = EvalExpr(&(p->pos), &val, p);
@@ -459,8 +528,15 @@ int ParseChar(ParsePtr p, int *err, int peek)
 	    DestroyParser(p);
 	    return 0;
 	}
+        while(*p->pos && (isempty(*p->pos))) {
+            p->pos++;
+        }
 	if (*p->pos != END_OF_EXPR) {
-	    *err = E_MISS_END;
+            if (*p->pos) {
+                *err = E_PARSE_ERR;
+            } else {
+                *err = E_MISS_END;
+            }
 	    DestroyParser(p);
 	    DestroyValue(val);
 	    return 0;
@@ -575,6 +651,57 @@ int ParseIdentifier(ParsePtr p, DynamicBuffer *dbuf)
     }
 }
 
+/***************************************************************/
+/*                                                             */
+/* ParseExpr                                                   */
+/*                                                             */
+/* We are expecting an expression here.  Parse it and return   */
+/* the value node tree.                                        */
+/*                                                             */
+/***************************************************************/
+expr_node * ParseExpr(ParsePtr p, int *r)
+{
+
+    int bracketed = 0;
+    expr_node *node;
+
+    if (p->isnested) {
+        *r = E_PARSE_ERR;  /* Can't nest expressions */
+        return NULL;
+    }
+    if (!p->pos) {
+        *r = E_PARSE_ERR;      /* Missing expression */
+        return NULL;
+    }
+
+    while (isempty(*p->pos)) (p->pos)++;
+    if (!*(p->pos)) {
+        *r = E_EOLN;
+        return NULL;
+    }
+    if (*p->pos == BEG_OF_EXPR) {
+	(p->pos)++;
+	bracketed = 1;
+    }
+    node = parse_expression(&(p->pos), r, NULL);
+    if (*r) {
+        return free_expr_tree(node);
+    }
+
+    if (bracketed) {
+	if (*p->pos != END_OF_EXPR) {
+            if (*p->pos) {
+                *r = E_PARSE_ERR;
+            } else {
+                *r = E_MISS_END;
+            }
+            return free_expr_tree(node);
+        }
+	(p->pos)++;
+    }
+    return node;
+}
+
 /***************************************************************/
 /*                                                             */
 /* EvaluateExpr                                                */
@@ -586,21 +713,22 @@ int ParseIdentifier(ParsePtr p, DynamicBuffer *dbuf)
 int EvaluateExpr(ParsePtr p, Value *v)
 {
 
-    int bracketed = 0;
     int r;
+    int nonconst = 0;
+    expr_node *node = ParseExpr(p, &r);
 
-    if (p->isnested) return E_PARSE_ERR;  /* Can't nest expressions */
-    while (isempty(*p->pos)) (p->pos)++;
-    if (!p->pos) return E_PARSE_ERR;      /* Missing expression */
-    if (*p->pos == BEG_OF_EXPR) {
-	(p->pos)++;
-	bracketed = 1;
+    if (r != OK) {
+        return r;
     }
-    r = EvalExpr(&(p->pos), v, p);
+    if (!node) {
+        return E_SWERR;
+    }
+
+    r = evaluate_expression(node, NULL, v, &nonconst);
+    free_expr_tree(node);
     if (r) return r;
-    if (bracketed) {
-	if (*p->pos != END_OF_EXPR) return E_MISS_END;
-	(p->pos)++;
+    if (nonconst) {
+        p->nonconst_expr = 1;
     }
     return OK;
 }
@@ -636,36 +764,36 @@ void Wprint(char const *fmt, ...)
 void Eprint(char const *fmt, ...)
 {
     va_list argptr;
+    char const *fname;
 
     /* Check if more than one error msg. from this line */
     if (!FreshLine && !ShowAllErrors) return;
 
-    if (FreshLine && FileName) {
-	FreshLine = 0;
-	if (strcmp(FileName, "-")) {
-	    (void) fprintf(ErrFp, "%s(%d): ", FileName, LineNo);
-            if (print_callstack(ErrFp)) {
-                (void) fprintf(ErrFp, ": ");
-            }
-        } else {
-	    (void) fprintf(ErrFp, "-stdin-(%d): ", LineNo);
-            if (print_callstack(ErrFp)) {
-                (void) fprintf(ErrFp, ": ");
-            }
-        }
-	if (DebugFlag & DB_PRTLINE) OutputLine(ErrFp);
-    } else if (FileName) {
-	fprintf(ErrFp, "       ");
-        if (print_callstack(ErrFp)) {
-            (void) fprintf(ErrFp, ": ");
-        }
+    if (!FileName) {
+        return;
     }
 
+    if (strcmp(FileName, "-")) {
+        fname = FileName;
+    } else {
+        fname = "-stdin-";
+    }
+    if (FreshLine) {
+        (void) fprintf(ErrFp, "%s(%d): ", fname, LineNo);
+    } else {
+	fprintf(ErrFp, "       ");
+    }
     va_start(argptr, fmt);
     (void) vfprintf(ErrFp, fmt, argptr);
     (void) fputc('\n', ErrFp);
     va_end(argptr);
-    return;
+    if (print_callstack(ErrFp)) {
+        (void) fprintf(ErrFp, "\n");
+    }
+    if (FreshLine) {
+        if (DebugFlag & DB_PRTLINE) OutputLine(ErrFp);
+    }
+    FreshLine = 0;
 }
 
 /***************************************************************/
@@ -834,6 +962,7 @@ int DoIf(ParsePtr p)
 	    }
     }
 
+    IfLinenos[NumIfs] = LineNo;
     NumIfs++;
     IfFlags &= ~(IF_MASK << (2*NumIfs - 2));
     IfFlags |= syndrome << (2 * NumIfs - 2);
@@ -898,7 +1027,7 @@ int DoIfTrig(ParsePtr p)
     if ((size_t) NumIfs >= IF_NEST) return E_NESTED_IF;
     if (ShouldIgnoreLine()) syndrome = IF_TRUE | BEFORE_ELSE;
     else {
-	if ( (r=ParseRem(p, &trig, &tim, 1)) ) return r;
+	if ( (r=ParseRem(p, &trig, &tim)) ) return r;
 	if (trig.typ != NO_TYPE) return E_PARSE_ERR;
 	dse = ComputeTrigger(trig.scanfrom, &trig, &tim, &r, 1);
 	if (r) {
@@ -1016,6 +1145,12 @@ int DoDebug(ParsePtr p)
 	    else     DebugFlag &= ~DB_ECHO_LINE;
 	    break;
 
+        case 's':
+        case 'S':
+	    if (val) DebugFlag |=  DB_PARSE_EXPR;
+	    else     DebugFlag &= ~DB_PARSE_EXPR;
+	    break;
+
 	case 'x':
 	case 'X':
 	    if (val) DebugFlag |=  DB_PRTEXPR;
@@ -1081,7 +1216,7 @@ int DoBanner(ParsePtr p)
 	}
     }
     DBufFree(&Banner);
-    
+
     err = DBufPuts(&Banner, DBufValue(&buf));
     DBufFree(&buf);
     return err;
@@ -1093,7 +1228,6 @@ int DoBanner(ParsePtr p)
 /*                                                             */
 /*  Enable or disable the RUN command under program control    */
 /*                                                             */
-/*                                                             */
 /***************************************************************/
 int DoRun(ParsePtr p)
 {
@@ -1120,6 +1254,38 @@ int DoRun(ParsePtr p)
     return VerifyEoln(p);
 }
 
+/***************************************************************/
+/*                                                             */
+/*  DoExpr                                                     */
+/*                                                             */
+/*  Enable or disable expression evaluation                    */
+/*                                                             */
+/***************************************************************/
+int DoExpr(ParsePtr p)
+{
+    int r;
+
+    DynamicBuffer buf;
+    DBufInit(&buf);
+
+    if ( (r=ParseToken(p, &buf)) ) return r;
+
+/* Only allow EXPR ON in top-level script */
+    if (! StrCmpi(DBufValue(&buf), "ON")) {
+	if (TopLevel()) ExpressionEvaluationDisabled = 0;
+    }
+/* But allow EXPR OFF anywhere */
+    else if (! StrCmpi(DBufValue(&buf), "OFF"))
+        ExpressionEvaluationDisabled = 1;
+    else {
+	DBufFree(&buf);
+	return E_PARSE_ERR;
+    }
+    DBufFree(&buf);
+
+    return VerifyEoln(p);
+}
+
 /***************************************************************/
 /*                                                             */
 /*  DoFlush                                                    */
@@ -1666,7 +1832,9 @@ System(char const *cmd, int is_queued)
     pid_t kid;
     int fd;
     int status;
+    int do_exit = 0;
     if (is_queued && IsServerMode()) {
+        do_exit = 1;
         /* Server mode... redirect stdin and stdout to /dev/null */
         kid = fork();
         if (kid == (pid_t) -1) {
@@ -1688,14 +1856,17 @@ System(char const *cmd, int is_queued)
             }
         } else {
             /* In the parent */
-            while (waitpid(kid, &status, 0) != kid) {
-                continue;
-            }
+            while (waitpid(kid, &status, 0) != kid) /* continue */ ;
             return;
         }
     }
     /* This is the child process or original if we never forked */
     r = system(cmd);
+    if (do_exit) {
+        /* In the child process, so exit! */
+        exit(0);
+    }
+
     if (r == 0) {
         return;
     }
