@@ -13,6 +13,18 @@
 /* Suppress unused variable warnings */
 #define UNUSED(x) (void) x
 
+#ifdef HAVE_STRDUP
+#define StrDup strdup
+#endif
+
+#ifdef HAVE_STRNCASECMP
+#define StrinCmp strncasecmp
+#endif
+
+#ifdef HAVE_STRCASECMP
+#define StrCmpi strcasecmp
+#endif
+
 /* Define a string assignment macro - be careful!!! */
 #define STRSET(x, str) { if (x) free(x); (x) = StrDup(str); }
 
@@ -31,22 +43,30 @@
 int CallUserFunc (char const *name, int nargs, ParsePtr p);
 int DoFset (ParsePtr p);
 int DoFunset (ParsePtr p);
+void UnsetAllUserFuncs(void);
 void ProduceCalendar (void);
 char const *SimpleTime (int tim);
 char const *CalendarTime (int tim, int duration);
 int DoRem (ParsePtr p);
 int DoFlush (ParsePtr p);
 void DoExit (ParsePtr p);
-int ParseRem (ParsePtr s, Trigger *trig, TimeTrig *tim, int save_in_globals);
+int ParseRem (ParsePtr s, Trigger *trig, TimeTrig *tim);
 int TriggerReminder (ParsePtr p, Trigger *t, TimeTrig *tim, int dse, int is_queued, DynamicBuffer *output);
 int ShouldTriggerReminder (Trigger *t, TimeTrig *tim, int dse, int *err);
 int DoSubst (ParsePtr p, DynamicBuffer *dbuf, Trigger *t, TimeTrig *tt, int dse, int mode);
 int DoSubstFromString (char const *source, DynamicBuffer *dbuf, int dse, int tim);
 int ParseLiteralDate (char const **s, int *dse, int *tim);
 int ParseLiteralTime (char const **s, int *tim);
+expr_node *parse_expression(char const **e, int *r, Var *locals);
+
+int evaluate_expression(expr_node *node, Value *locals, Value *ans, int *nonconst);
+int evaluate_expr_node(expr_node *node, Value *locals, Value *ans, int *nonconst);
+void print_expr_tree(expr_node *node, FILE *fp);
+void unlimit_execution_time(void);
+expr_node *free_expr_tree(expr_node *node);
 int EvalExpr (char const **e, Value *v, ParsePtr p);
 int DoCoerce (char type, Value *v);
-void PrintValue  (Value *v, FILE *fp);
+char const *PrintValue  (Value *v, FILE *fp);
 int CopyValue (Value *dest, const Value *src);
 int ReadLine (void);
 int OpenFile (char const *fname);
@@ -65,8 +85,9 @@ int JulianToGregorianOffset(int y, int m);
 int ParseChar (ParsePtr p, int *err, int peek);
 int ParseToken (ParsePtr p, DynamicBuffer *dbuf);
 int ParseIdentifier (ParsePtr p, DynamicBuffer *dbuf);
+expr_node * ParseExpr(ParsePtr p, int *r);
+void print_expr_nodes_stats(void);
 int EvaluateExpr (ParsePtr p, Value *v);
-int Evaluate (char const **s, Var *locals, ParsePtr p);
 int FnPopValStack (Value *val);
 void Eprint (char const *fmt, ...);
 void Wprint (char const *fmt, ...);
@@ -86,10 +107,11 @@ int VerifyEoln (ParsePtr p);
 int DoDebug (ParsePtr p);
 int DoBanner (ParsePtr p);
 int DoRun (ParsePtr p);
+int DoExpr (ParsePtr p);
 int DoErrMsg (ParsePtr p);
 int ClearGlobalOmits (void);
 int DoClear (ParsePtr p);
-int DestroyOmitContexts (void);
+int DestroyOmitContexts (int print_unmatched);
 int PushOmitContext (ParsePtr p);
 int PopOmitContext (ParsePtr p);
 int IsOmitted (int dse, int localomit, char const *omitfunc, int *omit);
@@ -103,13 +125,26 @@ int ComputeTrigger (int today, Trigger *trig, TimeTrig *tim, int *err, int save_
 int ComputeTriggerNoAdjustDuration (int today, Trigger *trig, TimeTrig *tim, int *err, int save_in_globals, int duration_days);
 int AdjustTriggerForDuration(int today, int r, Trigger *trig, TimeTrig *tim, int save_in_globals);
 char *StrnCpy (char *dest, char const *source, int n);
+
+#ifndef HAVE_STRNCASECMP
 int StrinCmp (char const *s1, char const *s2, int n);
+#endif
+
+#ifndef HAVE_STRDUP
 char *StrDup (char const *s);
+#endif
+
+#ifndef HAVE_STRCASECMP
 int StrCmpi (char const *s1, char const *s2);
+#endif
+
+void strtolower(char *s);
+
 Var *FindVar (char const *str, int create);
+SysVar *FindSysVar (char const *name);
 int DeleteVar (char const *str);
-int SetVar (char const *str, Value *val);
-int GetVarValue (char const *str, Value *val, Var *locals, ParsePtr p);
+int SetVar (char const *str, Value const *val);
+int GetVarValue (char const *str, Value *val);
 int DoSet  (Parser *p);
 int DoUnset  (Parser *p);
 int DoDump (ParsePtr p);
@@ -122,10 +157,10 @@ int DoMsgCommand (char const *cmd, char const *msg, int is_queued);
 int ParseNonSpaceChar (ParsePtr p, int *err, int peek);
 unsigned int HashVal (char const *str);
 int DateOK (int y, int m, int d);
-Operator *FindOperator (char const *name, Operator where[], int num);
-BuiltinFunc *FindFunc (char const *name, BuiltinFunc where[], int num);
+BuiltinFunc *FindBuiltinFunc (char const *name);
 int InsertIntoSortBuffer (int dse, int tim, char const *body, int typ, int prio);
 void IssueSortedReminders (void);
+UserFunc *FindUserFunc(char const *name);
 int UserFuncExists (char const *fn);
 void DSEToHeb (int dse, int *hy, int *hm, int *hd);
 int HebNameToNum (char const *mname);
@@ -174,15 +209,18 @@ int AddGlobalOmit(int dse);
 void set_lat_and_long_from_components(void);
 void set_components_from_lat_and_long(void);
 
+void DebugExitFunc(void);
+
 int GetTerminalBackground(void);
 
 char const *get_day_name(int wkday);
 char const *get_month_name(int mon);
 
-void set_cloexec(int fd);
+void set_cloexec(FILE *fp);
 int push_call(char const *filename, char const *func, int lineno);
 void clear_callstack(void);
 int print_callstack(FILE *fp);
+int have_callstack(void);
 void pop_call(void);
 void FixSpecialType(Trigger *trig);
 void WriteJSONTrigger(Trigger const *t, int include_tags, int today);
@@ -193,3 +231,11 @@ void WriteJSONTimeTrigger(TimeTrig const *tt);
 #include <wchar.h>
 void PutWideChar(wchar_t const wc, DynamicBuffer *output);
 #endif
+
+/* These functions are in utils.c and are used to detect overflow
+   in various arithmetic operators.  They have to be in separate
+   functions with extern linkage to defeat compiler optimizations
+   that would otherwise break the overflow checks. */
+extern int _private_mul_overflow(int a, int b);
+extern int _private_add_overflow(int a, int b);
+extern int _private_sub_overflow(int a, int b);
