@@ -17,6 +17,11 @@ static char const DontEscapeMe[] =
 #include "err.h"
 
 #include <string.h>
+
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+
 #include <stdio.h>
 #include <ctype.h>
 
@@ -46,6 +51,7 @@ char *StrnCpy(char *dest, char const *source, int n)
     return odest;
 }
 
+#ifndef HAVE_STRNCASECMP
 /***************************************************************/
 /*                                                             */
 /*  StrinCmp - compare strings, case-insensitive               */
@@ -64,6 +70,9 @@ int StrinCmp(char const *s1, char const *s2, int n)
     if (n) return (toupper(*s1) - toupper(*s2)); else return 0;
 }
 
+#endif
+
+#ifndef HAVE_STRDUP
 /***************************************************************/
 /*                                                             */
 /*  StrDup                                                     */
@@ -79,6 +88,9 @@ char *StrDup(char const *s)
     return ret;
 }
 
+#endif
+
+#ifndef HAVE_STRCASECMP
 /***************************************************************/
 /*                                                             */
 /*  StrCmpi                                                    */
@@ -98,6 +110,8 @@ int StrCmpi(char const *s1, char const *s2)
     return toupper(*s1) - toupper(*s2);
 }
 
+#endif
+
 /***************************************************************/
 /*                                                             */
 /*  DateOK                                                     */
@@ -116,6 +130,14 @@ int DateOK(int y, int m, int d)
     return 1;
 }
 
+void strtolower(char *s)
+{
+    while (*s) {
+        *s = tolower(*s);
+        s++;
+    }
+}
+
 /* Functions designed to defeat gcc optimizer */
 
 int _private_mul_overflow(int a, int b)
@@ -151,11 +173,15 @@ int _private_sub_overflow(int a, int b)
 int
 ShellEscape(char const *in, DynamicBuffer *out)
 {
-    while(*in) {
-        if (!strchr(DontEscapeMe, *in)) {
-            if (DBufPutc(out, '\\') != OK) return E_NO_MEM;
+    unsigned char const *i = (unsigned char const *) in;
+    while(*i) {
+        /* Don't escape chars with high bit set.  That will mangle UTF-8 */
+        if (! (*i & 0x80) ) {
+            if (!strchr(DontEscapeMe, *i)) {
+                if (DBufPutc(out, '\\') != OK) return E_NO_MEM;
+            }
         }
-        if (DBufPutc(out, *in++) != OK) return E_NO_MEM;
+        if (DBufPutc(out, *i++) != OK) return E_NO_MEM;
     }
     return OK;
 }
@@ -169,31 +195,33 @@ typedef struct cs_s {
 } cs;
 
 static cs *callstack = NULL;
+static cs *freecs = NULL;
 
 static void
 destroy_cs(cs *entry)
 {
-    if (entry->filename) free( (void *) entry->filename);
-    if (entry->func) free( (void *) entry->func);
-    free( (void *) entry);
+    entry->next = freecs;
+    freecs = entry;
 }
 
 
 int
 push_call(char const *filename, char const *func, int lineno)
 {
-    cs *entry = NEW(cs);
-    if (!entry) {
-        return E_NO_MEM;
+    cs *entry;
+    if (freecs) {
+        entry = freecs;
+        freecs = freecs->next;
+    } else {
+        entry = NEW(cs);
+        if (!entry) {
+            return E_NO_MEM;
+        }
     }
     entry->next = NULL;
-    entry->filename = StrDup(filename);
-    entry->func = StrDup(func);
+    entry->filename = filename;
+    entry->func = func;
     entry->lineno = lineno;
-    if (!entry->filename || !entry->func) {
-        destroy_cs(entry);
-        return E_NO_MEM;
-    }
     entry->next = callstack;
     callstack = entry;
     return OK;
@@ -215,11 +243,35 @@ clear_callstack(void)
 static void
 print_callstack_aux(FILE *fp, cs *entry)
 {
-    if (entry) {
-        print_callstack_aux(fp, entry->next);
-        fprintf(fp, "\n");
-        (void) fprintf(fp, "%s(%d): In function `%s'", entry->filename, entry->lineno, entry->func);
+    int i = 0;
+    char const *in = "In";
+    cs *prev = NULL;
+    while(entry) {
+        if (prev) {
+            in = "Called from";
+        }
+        if (!prev || strcmp(prev->func, entry->func) || strcmp(prev->filename, entry->filename) || prev->lineno != entry->lineno) {
+            if (prev) {
+                fprintf(fp, "\n");
+            }
+            (void) fprintf(fp, "    %s(%d): [#%d] %s function `%s'", entry->filename, entry->lineno, i, in, entry->func);
+        }
+        prev = entry;
+        entry = entry->next;
+        i++;
+        if (i > 10) {
+            break;
+        }
     }
+    if (entry) {
+        (void) fprintf(fp, "\n    [remaining call frames omitted]");
+    }
+}
+
+int
+have_callstack(void)
+{
+    return (callstack != NULL);
 }
 
 int
