@@ -22,7 +22,6 @@
 #include <errno.h>
 #include <locale.h>
 #include "types.h"
-#include "expr.h"
 #include "globals.h"
 #include "protos.h"
 #include "err.h"
@@ -39,8 +38,6 @@ static int IntMax = INT_MAX;
 
 static Var *VHashTbl[VAR_HASH_SIZE];
 
-typedef int (*SysVarFunc)(int, Value *);
-
 static double
 strtod_in_c_locale(char const *str, char **endptr)
 {
@@ -175,6 +172,19 @@ static int terminal_bg_func(int do_set, Value *val)
     return OK;
 }
 
+static int trig_time_func(int do_set, Value *val)
+{
+    UNUSED(do_set);
+    if (LastTriggerTime != NO_TIME) {
+        val->type = TIME_TYPE;
+        val->v.val = LastTriggerTime;
+    } else {
+        val->type = INT_TYPE;
+        val->v.val = 0;
+    }
+    return OK;
+}
+
 static int trig_date_func(int do_set, Value *val)
 {
     UNUSED(do_set);
@@ -309,6 +319,29 @@ static int datetime_sep_func(int do_set, Value *val)
     return OK;
 }
 
+static int expr_time_limit_func(int do_set, Value *val)
+{
+    if (!do_set) {
+        val->type = INT_TYPE;
+        val->v.val = ExpressionEvaluationTimeLimit;
+        return OK;
+    }
+    if (val->type != INT_TYPE) return E_BAD_TYPE;
+    if (val->v.val < 0) return E_2LOW;
+
+    if (!TopLevel()) {
+        /* Ignore attempts to set from non-toplevel unless it's
+           lower than current value */
+        if (val->v.val == 0 ||
+            val->v.val >= ExpressionEvaluationTimeLimit) {
+            return OK;
+        }
+    }
+
+    ExpressionEvaluationTimeLimit = val->v.val;
+    return OK;
+}
+
 static int default_color_func(int do_set, Value *val)
 {
     int col_r, col_g, col_b;
@@ -477,10 +510,10 @@ int DeleteVar(char const *str)
 /*                                                             */
 /*  SetVar                                                     */
 /*                                                             */
-/*  Set the indicate variable to the specified value.          */
+/*  Set the indicated variable to the specified value.         */
 /*                                                             */
 /***************************************************************/
-int SetVar(char const *str, Value *val)
+int SetVar(char const *str, Value const *val)
 {
     Var *v = FindVar(str, 1);
 
@@ -498,20 +531,10 @@ int SetVar(char const *str, Value *val)
 /*  Get a copy of the value of the variable.                   */
 /*                                                             */
 /***************************************************************/
-int GetVarValue(char const *str, Value *val, Var *locals, ParsePtr p)
+int GetVarValue(char const *str, Value *val)
 {
     Var *v;
 
-    /* Try searching local variables first */
-    v = locals;
-    while (v) {
-	if (! StrinCmp(str, v->name, VAR_NAME_LEN))
-	    return CopyValue(val, &v->v);
-	v = v->next;
-    }
-
-    /* Global variable... mark expression as non-constant */
-    if (p) p->nonconst_expr = 1;
     v=FindVar(str, 0);
 
     if (!v) {
@@ -544,6 +567,11 @@ int DoSet (Parser *p)
 	ParseNonSpaceChar(p, &r, 0);
     }
 
+    if (p->isnested) {
+        Eprint("%s", "Do not use [] around expression in SET command");
+        return E_CANTNEST_FDEF;
+    }
+
     r = EvaluateExpr(p, &v);
     if (r) {
 	DBufFree(&buf);
@@ -772,16 +800,6 @@ int DoPreserve (Parser *p)
 /*                                                             */
 /***************************************************************/
 
-/* The structure of a system variable */
-typedef struct {
-    char const *name;
-    char modifiable;
-    int type;
-    void *value;
-    int min; /* Or const-value */
-    int max;
-} SysVar;
-
 /* Macro to access "min" but as a constval.  Just to make source more
    readable */
 #define constval min
@@ -818,6 +836,7 @@ static SysVar SysVarArr[] = {
     {"DontTrigAts",    0,  INT_TYPE,     &DontIssueAts,        0,      0 },
     {"EndSent",        1,  STR_TYPE,     &EndSent,             0,      0 },
     {"EndSentIg",      1,  STR_TYPE,     &EndSentIg,           0,      0 },
+    {"ExpressionTimeLimit", 1, SPECIAL_TYPE, expr_time_limit_func, 0,  0 },
     {"February",       1,  STR_TYPE,     &DynamicMonthName[1], 0,      0 },
     {"FirstIndent",    1,  INT_TYPE,     &FirstIndent,         0,      132 },
     {"FoldYear",       1,  INT_TYPE,     &FoldYear,            0,      1 },
@@ -887,6 +906,7 @@ static SysVar SysVarArr[] = {
     {"Tm",             0,  SPECIAL_TYPE, trig_mon_func,        0,      0 },
     {"Today",          1,  STR_TYPE,     &DynamicToday,        0,      0 },
     {"Tomorrow",       1,  STR_TYPE,     &DynamicTomorrow,     0,      0 },
+    {"Tt",             0,  SPECIAL_TYPE, trig_time_func,       0,      0 },
     {"Tuesday",        1,  STR_TYPE,     &DynamicDayName[1],   0,      0 },
     {"Tw",             0,  SPECIAL_TYPE, trig_wday_func,       0,      0 },
     {"Ty",             0,  SPECIAL_TYPE, trig_year_func,       0,      0 },
@@ -905,7 +925,6 @@ static SysVar SysVarArr[] = {
 };
 
 #define NUMSYSVARS ( sizeof(SysVarArr) / sizeof(SysVar) )
-static SysVar *FindSysVar (char const *name);
 static void DumpSysVar (char const *name, const SysVar *v);
 /***************************************************************/
 /*                                                             */
@@ -1002,7 +1021,7 @@ int GetSysVar(char const *name, Value *val)
 /* Find a system var with specified name.                      */
 /*                                                             */
 /***************************************************************/
-static SysVar *FindSysVar(char const *name)
+SysVar *FindSysVar(char const *name)
 {
     int top=NUMSYSVARS-1, bottom=0;
     int mid=(top + bottom) / 2;
